/*
  chil.h
  Header file for libchil, a library for reading Chilbolton-format radar data
  Copyright (C) 1999, 2000  Robin Hogan
*/

/* 
   First the preprocessor definitions -
   Many were taken from various header files used by Mark Blackman's rad93.c
*/

#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif                          /* __cplusplus */

#ifdef DEBUGLINE
#define DEBUG_LINE CHILPrintf("DEBUG: %s got to line %d\n",__FILE__,__LINE__);
#else
#define DEBUG_LINE /* */
#endif

#define CHIL_LAT  51.1445
#define CHIL_LON  -1.4370
#define CHIL_ALT  90.0

#define SINGAPORE_LAT  1.3423
#define SINGAPORE_LON  103.6806
#define SINGAPORE_ALT  49.0

/* System IDs: these values are contained in byte 29 of the IR in some
   recent data */
#define UNKNOWN_SYSTEM 0
#define CAMRA 1
#define RABELAIS 2
#define GALILEO 3
#define SINGAPORE 4

/* System frequencies */
#define CAMRA_FREQ 3.075 /* tx freq, Ghz */		  
#define RABELAIS_FREQ 34.960 /* number from official documents */
#define GALILEO_FREQ 94.00      /* approximately... */
#define SINGAPORE_FREQ 3.0   /* approximately */

/* Parameter IDs */
#define CAMRA_ZH	1
#define CAMRA_ZDR	2
#define CAMRA_LDR	3
#define CAMRA_V 	4
#define CAMRA_WIDTH	5
#define CAMRA_KDP	6
#define CAMRA_PHIDP	7
#define CAMRA_VH	8
#define CAMRA_VV	9
#define CAMRA_V2	10
#define CAMRA_WIDTHH	11
#define CAMRA_WIDTHV	12
#define RABELAIS_ZH	30
#define RABELAIS_ZDR	31
#define RABELAIS_PHIDP	32
#define RABELAIS_V	33
#define RABELAIS_VV	34
#define SINGAPORE_ZH	40
#define SINGAPORE_LDR	41
#define SINGAPORE_V	42
#define SINGAPORE_WIDTH	43
#define GALILEO_ZH	50
#define GALILEO_V	51

/* scan types */
#define RHI_93 1
#define PPI_93 2
#define SP_93 3
#define FIXED_93 4
#define CAL_93 5
#define TS_93 6

#define UF_CAL 0
#define UF_PPI 1
#define UF_COPLANE 2
#define UF_RHI 3
#define UF_VERT 4
#define UF_DWELL 5
#define UF_MANUAL 6
#define UF_IDLE 7

/* New format bitwise parameter IDs and corresponding masks*/
#define UNKNOWN 0
#define ZH 1
#define ZDR 2
#define LDR 3
#define PHIDP 4
#define V1 5
#define V2 6
#define WIDTH_H 7
#define WIDTH_V 8
#define WIDTH 7
#define V_STD 15

#define ZH_MASK (1L<<0)
#define ZDR_MASK (1L<<1)
#define LDR_MASK (1L<<2)
#define PHIDP_MASK (1L<<3)
#define V1_MASK (1L<<4)
#define V2_MASK (1L<<5)
#define WIDTH_MASK (1L<<6)
#define OVERFLOW_MASK (1L<<7)

#define UNKNOWN_NAME	"Unknown"
#define ZH_NAME 	"Zh"
#define ZDR_NAME	"Zdr"
#define LDR_NAME	"Ldr"
#define PHIDP_NAME	"phi_dp"
#define VEL1_NAME	"v1"
#define VEL2_NAME	"v2"
#define WIDTH_NAME	"width"      
#define WIDTH_H_NAME	"width_h"      
#define WIDTH_V_NAME	"width_v"      
#define ZH1_NAME	"Zh"
#define ZH2_NAME	"Zh2"
#define Z_NAME          "Z"
#define VEL_NAME	"v"
#define VELSTD_NAME	"sigma_v"

#define UNKNOWN_UNITS	"unknown"
#define ZH_UNITS	"dBZ"
#define ZDR_UNITS	"dB"
#define LDR_UNITS	"dB"
#define PHIDP_UNITS	"degrees"
#define VEL1_UNITS	"m s-1"
#define VEL2_UNITS	"m s-1"
#define WIDTH_UNITS	"m s-1"      
#define ZH1_UNITS	"dBZ"
#define ZH2_UNITS	"dBZ"
#define VEL_UNITS	"m s-1"
#define VELSTD_UNITS	"m s-1"

#define VEL_UNITS_HTML  "m s<sup>-1</sup>"
#define PHIDP_UNITS_HTML "&deg;"

/* Scalings */
#define ZH_SCALING_93  0.25
#define ZDR_SCALING_93 0.125
#define LDR_SCALING_93 0.25
#define ZH_SCALING_HP  0.01
#define ZDR_SCALING_HP 0.01
#define PHASE_SCALING_93 0.001
#define MIN_SPACING 37.5
#define MIN_SPACING_GALILEO_NEW 30.0

/* In the coherent processing system, no signal is recorded as -199.0 dBZ */
#define MIN_ZH -190.0

/* mb. 27/07/95 ....... all according to kev's document of oct. 1994 */
#define END_OF_SCAN 65281
#define END_OF_RASTER 61440
#define END_OF_RASTER_B 61540
#define RASTER_ABORTED 65535

/* Other codes which I think are appropriate for raw files */
#define STATUS_OK 30720
#define STATUS_ABORTED 32767
#define STATUS_NC 32641 /* Aborted FIXed? Or Not Complete raster?*/

/* for ChilboltonFile structure element `position', and some return values */
#define END_OF_DATA -1
#define FILE_CLOSED 0
#define START_OF_IR 1
#define START_OF_DATA 2
#define START_OF_RAY 3
#define START_OF_FILE 4

/*
  for noise subtraction. These values are SYSTEM dependent -
  here appropriate for the galileo. NO LONGER USED
*/
/* #define MIN_V_STD 0.05 */
/* #define MAX_V_STD 2.5 */

#define CHIL_BLOCKSIZE 4092
#define CHIL_RAYBUFFERSIZE 4096

#ifndef uint
#define uint unsigned int
#endif
#ifndef uword
#define uword unsigned short int
#endif
#ifndef word
#define word short int
#endif
#ifndef uchar
#define uchar unsigned char
#endif

#define MAX_FIELDS 15
#define MAX_FIELDNAME_LENGTH 6
#define MAX_FILENAME_LENGTH 512
#define MAX_FILES 8
#define MAX_RASTERS 1000

#define CHIL_OK 0
#define CHIL_ERROR -1
#define CHIL_NOTFOUND -2
#define CHIL_NOTDONE -3
#define CHIL_ENDOFBLOCK 1
#define CHIL_ENDOFDATA 2

extern void (*CHILFree) (void *);
extern void* (*CHILCalloc) (size_t nelems, size_t elsize);
extern void* (*CHILRealloc) (void *ptr, size_t nbytes);
extern int (*CHILPrintf) (const char *, ...);

extern float chil_no_signal_value;
extern char secondary_sum_dir[MAX_FILENAME_LENGTH];
extern uchar chil_maxfields; /* set by `-maxfields' in chil_api.c */

/* Now for some useful structures */

typedef struct {
  char *name[MAX_FILES];
  char **allnames;
  uword first[MAX_FILES],last[MAX_FILES];
  word n;
  char buffer[MAX_FILES*MAX_FILENAME_LENGTH];
  char allrasters[MAX_FILES];
} FileList;

typedef struct {
  FILE *file;
  float Zcal;           /* Calibration figure in summary file (dB) */
  float ZDRcal;         /* Calibration figure in summary file (dB) */
  int bytes_read;	/* bytes of current raster read */
  uword iraster;	/* the index of the current raster in the file,
			   starting at 0 */
  uword iscan;		/* index of the current scan */
  /* the following are read from the summary file - 
     undefined if it can't be found */
  uword nrasters;	/* number of rasters in the file */
  uword raster[MAX_RASTERS];	/* array of raster numbers in the order they appear */
  uword block[MAX_RASTERS];	/* corresponding block locations of each raster */
  uword day[MAX_RASTERS];
  uword month[MAX_RASTERS];
  uword year[MAX_RASTERS];
  uchar is_raw;		/* flag to indicate that the file is raw - 
			   i.e. contains one raster with the
			   normal big-endian byte order reversed */
  uchar summary_exists;	/* indicates that a summary file has been found 
			   and read */
  uchar cals_constant;  /* are sum-file calibration figures constant with raster? */
  char name[MAX_FILENAME_LENGTH];
  char default_sum_dir[MAX_FILENAME_LENGTH];
  char position;
} ChilboltonFile;

typedef struct {
  float frequency;
  float (*field_function[MAX_FIELDS]) (short);
  uword pulse_pairs;
  uword min_gate;	/* minimum gate, def: 17 */
  uword max_gate;	/* maximum gate, def: 273 */
  uword day;		/* day of month scan was taken */
  uword month;		/* month of scan */
  uword year;		/* year of scan */
  uword file;		/* file number, (5...), like old tape numbers */
  word threshold;	/* threshold for (displayed ??) Zh in dbZ.
			   def: -32, 3gig only */
  word min_angle;	/* min. angle in raster (arcmin), elevation
			   for RHI */
  uword max_angle;	/* max. angle in raster (arcmin), azimuth for
			   PPI */
  uword central;	/* central angle for (arcmin) for coord.
			   transform */
  uword raster;		/* raster number of this scan */
  word init_angle;	/* initial azim/elev of scan (arcmin) */
  word angle_inc;	/* angle increment between scans (arcmin)
			   ST?? */
  uword gate_spacing;	/* gate spacing in metres */
  word range_offset;	/* gate offset in metres, from A scope */
  uchar raw[64];	/* raw information record data */
  uchar recording_flag;	/* >0, MT on (recording) */
  uchar spacing_raw;
  uchar spacing_ave;
  uchar nrawfields;
  uchar field_id[MAX_FIELDS]; /* ID code for each data field, 1 = Zh, 2 =
				 Zdr, 3 = Ldr, 4 = V, 5 = deltaV, 6 = Kdp,
				 7 = Phidp, 8 = Vh, 9 = Vv, 10=Vhv ... */
  uchar field_type[MAX_FIELDS]; /* similar to above but system-independent */
  uchar nscans;		/* no. of scans in raster */
  uchar expt_id;	/* experiment identifier */
  uchar scan_type;	/* scan type,
			   1 = rhi, 2 = ppi,3 = sp,4 = fixed,
			   5 = calibration, 6 = time series */
  uchar sweep_mode;	/* sweep_mode, universal format */
  uchar sens_flag;	/* 1 = 20db of atten., 0..no atttenuation */
  uchar hour;		/* hour of raster start */
  uchar oper[3];	/* 3 ASCII bytes identifying operater */
  uchar minute;		/* min. of raster start */
  uchar second;		/* second of raster start */
  uchar cmpl_scans;	/* scans in raster completed (first scan = 1) */
  uchar system_id;
  uchar range_normalised;
  uchar time_stamping;
  uchar cloud_data;
  char scan_velocity ;	/* scan velocity = deg/sec*50.0 */
} RasterInfo;

typedef struct {
  long int maxsize;
  double *sum[MAX_FIELDS];
  double *sum2[MAX_FIELDS]; /* To correctly average v we need to sum I & Q
			       components separately */
  float range_offset;   /* not the same as the RasterInformation version
			   if gates are averaged */
  float *range;		/* WITHOUT applying range correction */
  float *data[MAX_FIELDS];
  int *npixels;         /* Number if pixels so far averaged at each height */
  float elevation, azimuth; /* in degrees */
  float utc;		/* in decimal hours */
  float raw_utc;	/* in decimal hours */
  float last_utc;	/* necessary for old cloud time stamping via
			   az and el */
  float sum_elevation, sum_azimuth; /* in degrees */
  float sum_utc;	/* in decimal hours */
  float v_floor;	/* rejection threshold for v in dB above the 
			   rejection threshold for Z */
  float sigma_v_floor;	/* rejection threshold for sigma_v in dB above the 
			   rejection threshold for Z */
  float nez;		/* Noise-equivalent reflectivity at 1 km; 
			   =0 if no processing */
  float v1fold, v2fold;
  int field_info[MAX_FIELDS]; /* some of the derived fields need somewhere to 
				 store the number of the parameter they are 
				 derived from (for instance) */
  int zh_index;         /* Output of CHILFindField(ray, ZH); */
  uword startgate, endgate, ngates, avegates, averays, index;
  uword day, month, year; /* These change within a raster so can be useful */
  uword raw_day, raw_month, raw_year; /* of non-averaged rays */
  word hour_mod;	/* necessary for old cloud time stamping via 
			   az and el */
  word buffer[CHIL_RAYBUFFERSIZE];
  uchar nrawfields;	/* the number of raw fields stored in each ray */
  uchar ntotal;		/* the total number of fields, both raw and derived */
  uchar nfields;	/* the number of fields requested */
  uchar field_type[MAX_FIELDS]; /* for both raw and derived fields */
  uchar range_normalised;
  uchar correct_range;
  char use_v1fold, use_v2fold;
  char pci_clean;       /* same as the value in the ChilboltonData structure */
} Ray;

typedef struct {
  /* Floats */
  float noisefactor, rangeoffset;
  float Zcal, ZDRcal, LDRcal, PHIDPcal;
  float sigma_v_floor, v_floor;
  float nez_Zh, nez_Zv, nez_Zhv, maxLDR;
  float threshold, dopplerthreshold, reject_val, minelev;
  float noclutter_minv;
  float v1fold, v2fold;
  /* Ints */
  int buflen, iray, iblock, nblock, datelen;
  int argc, optstart;
  int file;
  int extragates;
  uint noisegates;
  uint reqday, reqmonth, reqyear;
  uint clean_cycle, clean_offset, clean_gates;
  /* Pointers */
  char *filebuf;
  char **argv;
  /* Structures */
  ChilboltonFile chil_file;
  RasterInfo raster_info;
  Ray ray;
  FileList file_list;
  /* Short ints */
  uword raster, scan, avegates, averays, reject_gate;
  uword noclutter_maxgate;
  uchar maxfields;
  /* Flags */
  char firsttime, process, clean, noclutter, newrangeoffset;
  char polarprocess, dopplerprocess;
  char warn_cloudnoise, warn_polarnoise;
  char warn_clutterLdr, warn_oscillator;
  char warn_minelev, warn_confused;
  char warn_dopplernoise, warn_glitch, warn_vfold, warn_calchange;
  char nobreak, use_v1fold, use_v2fold;
  char verbose, quiet, progress, rangecorrection, reject;
  char reversev, custom_clean, pci_clean;
  char forceZcal, forceZDRcal, forceLDRcal, forcePHIDPcal;
  char autocal;         /* apply calibration factors in the summary file */
  char velocity_away;
} ChilboltonData;

typedef struct {
  word azimuth;		/* azimuth of ray in arcminutes */
  word elevation;	/* elevation of ray in arcminutes */
} RayHeader;

typedef struct {
  uword mins_since_midnight;
  uword centisecs;	/* since last minute */
} TimeStamp;



/* *** PROTOTYPES *** */

/* simple interface */
void CHILUsage(ChilboltonData *);
void CHILOptions(ChilboltonData *);
int CHILInit(int argc, char **argv, ChilboltonData *);
int CHILAdvance(ChilboltonData *);
int CHILFetchRay(ChilboltonData *);
void CHILEnd(ChilboltonData *);
char *CHILGetOptions(ChilboltonData *);
char *CHILFieldInformation(const ChilboltonData *, int field);
char *CHILFileInformation(ChilboltonData *);
/* functions for looking at summary files */
int CHILGetFileList(uword day,uword month,uword year,int Argc,char **Argv,FileList * filelist);
FILE *CHILOpenSummary(char * datfilename);
/* functions used to open and close radar files: */
int CHILOpen(char *,ChilboltonFile *);
int CHILClose(ChilboltonFile *);
char *CHILSecondarySumDir(char * sec_sum_dir);
/* funtions used to negotiate radar files: */
int CHILReadInformationRecord(ChilboltonFile *,RasterInfo *);
int CHILNextRaster(ChilboltonFile *,RasterInfo *);
int CHILSeekRaster(ChilboltonFile *,uword required_raster);
int CHILSkipScans(ChilboltonFile *,RasterInfo *,int nscans);
int CHILSkipScansAndCount(ChilboltonFile * chilfile, RasterInfo * info,
			  int *pscans);
int CHILSkipRay(ChilboltonFile *,RasterInfo *);
/* ray functions */
int __CHILSetRay(const RasterInfo *, int avegates, int averays, Ray *, uchar isnew);
#define CHILInitRay(info, avegates, averays, ray) \
  __CHILSetRay(info, avegates, averays, ray, 1)
#define CHILReInitRay(info, avegates, averays, ray) \
  __CHILSetRay(info, avegates, averays, ray, 0)
int CHILSetFoldingVelocities(Ray *, float v1fold, float v2fold);
#define CHILSetFoldingVelocity(ray, vfold) \
  CHILSetFoldingVelocities(ray, vfold, vfold)
int CHILFreeRay(Ray *);
int CHILReadRawRay(ChilboltonFile *,RasterInfo *,Ray *);
int CHILReadRay(ChilboltonFile *,RasterInfo *,Ray *);
int CHILNextRay(ChilboltonFile *,const RasterInfo *);
int CHILSubtractNoise(Ray *ray,int noisegates,float factor);
int CHILSubtractOscillatorNoise(Ray *ray,int noisegates,float factor);
int CHILIsCorrupt(Ray *ray, uword gate, uchar field, float maxval);
int CHILIsRangeGlitched(Ray *ray, uword mingate, float maxjump);
int CHILIsPCIGlitched(RasterInfo *info, Ray *ray);
int CHILCleanRay(Ray *ray);
int CHILDoubleCleanRay(Ray *ray);
int CHILCustomCleanRay(Ray *ray, uint cycle, uint offset, uint cleangates);
int CHILSuppressClutter(Ray *ray, uword max_gate, float min_v);
int CHILRangeCorrection(RasterInfo *,Ray *ray);
int CHILCalibrateZh(Ray *ray,float Zcal);
int CHILCalibrateField(Ray *ray, int fieldid, float cal);
int CHILSubtractNoisePolar(Ray *ray, float nezh, float nezv, float nezhv, float threshold);
int CHILSuppressClutterWithLdr(Ray *ray, float maxldr);
int CHILRemoveNoiseDoppler(Ray *ray, float threshold);
int CHILReverseVelocities(Ray *ray);
/* debugging functions */
int CHILPrintFileInformation(ChilboltonFile *);
int CHILPrintRasterInformation(RasterInfo *);
int CHILPrintRayInformation(Ray * ray);
int CHILPrintFieldInformation(Ray * ray);
int CHILPrintInformationRecord(RasterInfo *);
/* information functions: */
char *CHILFieldDescription(const Ray *ray,int field,char * desc);
char *CHILFieldName(const Ray *ray,int field,char * name);
char *CHILFieldUnits(const Ray *ray,int field,char * units);
char *CHILFieldUnitsHTML(const Ray *ray,int field,char * units);
char *CHILSystemName(const RasterInfo *info,char *name);
char *CHILSystemInformation(const RasterInfo *info, char *out);
char *CHILLongSystemName(const RasterInfo *info,char *name);
char *CHILScanType(const RasterInfo *info, char *type);
char *CHILExperimentCode(const RasterInfo *info, char *code);
char *CHILLocation(const RasterInfo *info);
float CHILAntennaDiameter(const RasterInfo *info);
float CHILLatitude(const RasterInfo *info);
float CHILLongitude(const RasterInfo *info);
float CHILAltitude(const RasterInfo *info);
/* conversion functions */
float CHILraw(short);
float CHILzh(short);
float CHILzh_highprecision(short);
float CHILzdr(short);
float CHILzdr_highprecision(short);
float CHILldr(short);
float CHILphidp_old(short);
float CHILphidp(short);
float CHILphidp_new(short x);
float CHILvel_old(short);
float CHILvel(short);
/* utilities */
void CHILIncrDate(uword *day,uword *month,uword *year);
void CHILDecrDate(uword *day,uword *month,uword *year);
int CHILGetDate(char *datestr, Ray *ray);
uword CHILTrueYear(uword falseyear);
int CHILFindField(Ray *ray, int field);
#define CHILFindZH(ray) ((ray)->zh_index)
/* memory allocation functions */
void CHILFreeDefault (void *ptr);
void *CHILCallocDefault (size_t nelems, size_t elsize);
void *CHILReallocDefault (void *ptr, size_t nbytes);
void CHILSetFreeFunction(void (*free_function) (void *));
void CHILSetCallocFunction(void* (*calloc_function) (size_t, size_t));
void CHILSetReallocFunction(void* (*realloc_function) (void *, size_t));
/* message reporting */
int CHILPrintfDefault(const char *format, ...);
void CHILSetPrintfFunction(int (*printf_function) (const char *format, ...));
/* miscellaneous functions */
void CHILSetNoSignalValue(float no_signal_value);
/* accessing calibration table */
int CHILReadCalibrationTable(char *filename);
int CHILGetCalibrationValue(int year, int month, int day, float *cal);
/* Creating filenames */
int CHILCreateFileName(char *dest, char *format, char *prefix,
		       uword file, uword raster, uword scan,
		       uword year, uword month, uword day,
		       uword hour, uword minute, uword second,
		       char *system, char *scantype);
void CHILCreateFileNameUsage();

#ifdef __cplusplus
}                               /* extern "C" */
#endif                          /* __cplusplus */


