/*
  ceil_read.c

  Useful functions for accessing ceilometer data files
  All use the prefix LC, which stands for lidar ceilometer

  R J Hogan, 10 May 1999
*/

#include <math.h>
#include "ceil.h"

/* Must cope with dodgy ray - in this case jump to next \n followed by a `-', 
   then pop the `-' back on the file */
static
int 
_LCSkipRay(FILE * ceilfile)
{
  char c1,c2;

  c1='\0';
  while ((c2=getc(ceilfile)) != '-' && c1 != '\n' ) {
    if (c2 == EOF)
      return 0;
    c1=c2;
  }
  ungetc('-',ceilfile);
  return -1;
}

/* An internal function to get the cloud height information. This is called 
   by both LCGetFullRay and LCGetCloudBase */
static
int 
_LCGetCloudBasesOnly(FILE * ceilfile,LCTime * time,
		     LCCloudBases * cloudbases,int * p_loop)
{
  char buffer[LC_MAXLINE];
  int detection_status,range_code;
  long int units_code;
  int charsassigned = 0;

#ifdef OLD_BUGGY_VERSION
  const char file_head[]="-Ceilometer Logfile";
  const char file_tail[]="-File Closed: ";

  while ( ! fgets(buffer,LC_MAXLINE,ceilfile)) {
    if (feof(ceilfile)) {
      return 0;
    }
  }

  /* First check whether we're at the beginning or end of a file */
  if (strncasecmp(file_head,buffer,5) == 0) {
    fgets(buffer,LC_MAXLINE,ceilfile);
    fgets(buffer,LC_MAXLINE,ceilfile);
  }
  else if (strncasecmp(file_tail,buffer,5) == 0) {
    fgets(buffer,LC_MAXLINE,ceilfile);
    fgets(buffer,LC_MAXLINE,ceilfile);
    fgets(buffer,LC_MAXLINE,ceilfile);
    return 0;
  }

#if DEBUG
  printbuffer;
#endif

  /* Now extract the date */
  if ((charsassigned = sscanf(buffer,"-%4d-%2d-%2d %2d:%2d:%2d",
			      &(time->year),&(time->month),&(time->day),
			      &(time->hour),&(time->minute),&(time->second)))
      != 6) {
    fprintf(stderr, "%d ", charsassigned);
    return 0;
  }
  fprintf(stderr, "6");

#else
  /* New version which copes with dodgy files much better */
  /* We want a line that begins with "-" and contains a valid date */
  do {
    while (!fgets(buffer,LC_MAXLINE,ceilfile)) {
      if (feof(ceilfile)) {
	return 0;
      }
    }
  }
  while (sscanf(buffer,"-%4d-%2d-%2d %2d:%2d:%2d",
		&(time->year),&(time->month),&(time->day),
		&(time->hour),&(time->minute),&(time->second))
    < 6);

#endif


  /* get range flag from next line */
  fgets(buffer,LC_MAXLINE,ceilfile);
  if (sscanf(buffer,"%*7c%1x",&range_code) == 0)
    range_code=4;
  if (range_code == 4)
    *p_loop=24;
  else
    *p_loop=47;

#if DEBUG
  fprintf(stderr,"#-%c-%c-%c-%c ",buffer[0],buffer[1],buffer[6],buffer[7]);
  printbuffer;
#endif

  /* Get cloud base heights */
  fgets(buffer,LC_MAXLINE,ceilfile);


  cloudbases->height1 = cloudbases->height2 = cloudbases->height3 = 0;
  cloudbases->units = LC_UNKNOWN;
  cloudbases->number=sscanf(buffer,"%*3c%d %d %d",
			    &(cloudbases->height1),&(cloudbases->height2),
			    &(cloudbases->height3));
  if (sscanf(buffer,"%1x%*20c%8lx",&detection_status,&units_code) == 2) {
    if ((units_code & 256L) == 256L)
      cloudbases->units = LC_METRES;
    else {
#if FEET_TO_METRES
      cloudbases->height1 = LCFeetToMetres(cloudbases->height1);
      cloudbases->height2 = LCFeetToMetres(cloudbases->height2);
      cloudbases->height3 = LCFeetToMetres(cloudbases->height3);
      cloudbases->units = LC_METRES;
#else
      cloudbases->units = LC_FEET;
#endif    
    }
  }

  return 1;
}

/* Gets a ray of data - the three cloud bases plus the profile of beta */
int
LCGetFullRay(FILE *ceilfile,LCTime *time,LCCloudBases *cloudbases,
	     LCBackscatterProfile *profile)
{
  char buffer[LC_MAXLINE];
  int loop,i,j;
  short int *ibeta=profile->beta;

  if ( ! _LCGetCloudBasesOnly(ceilfile,time,cloudbases,&loop)) {
    return 0;
  }
  
  /* Read next line */
  fgets(buffer,LC_MAXLINE,ceilfile);

  /* Get elevation of instrument */
  buffer[27] = '\0';
  if (sscanf(buffer+23, "%f ", &profile->elevation) == 0) {
    profile->elevation = LC_NO_ELEVATION;
  }

  /* Read backscatter coefficient */
  for (i=0;i<loop;i++) {
    fgets(buffer,LC_MAXLINE,ceilfile);
    j=i*16;
    if (sscanf(buffer,"%*3d%4hX%4hX%4hX%4hX%4hX%4hX%4hX%4hX%4hX%4hX"
	       "%4hX%4hX%4hX%4hX%4hX%4hX\n",
	       ibeta+j,ibeta+j+1,ibeta+j+2,ibeta+j+3,
	       ibeta+j+4,ibeta+j+5,ibeta+j+6,ibeta+j+7,
	       ibeta+j+8,ibeta+j+9,ibeta+j+10,ibeta+j+11,
	       ibeta+j+12,ibeta+j+13,ibeta+j+14,ibeta+j+15) != 16)
      return _LCSkipRay(ceilfile);
  }

#if DEBUG
  printbuffer;
  fprintf(stderr,"25th point: %hd\n",profile->beta[25]);
#endif

  profile->ngates = loop*16;
  for (i=0;i<loop*16;i++)
    profile->range[i] = LC_GATE_SPACING * (i+1);

  /* Skip next two lines */
  fgets(buffer,LC_MAXLINE,ceilfile);
  fgets(buffer,LC_MAXLINE,ceilfile);

  return 1;

}

/* Gets the three cloud bases from the next ray but ignores the beta profile */
int
LCGetCloudBases(FILE *ceilfile,LCTime *time,LCCloudBases *cloudbases)
{
  char buffer[LC_MAXLINE];
  int loop,i;

  if ( ! _LCGetCloudBasesOnly(ceilfile,time,cloudbases,&loop))
    return 0;

  /* Skip remaining lines */
  for (i=0;i<loop+3;i++)
    fgets(buffer,LC_MAXLINE,ceilfile);

  return 1;

}

/* Read-in `nrays' rays and average into one */
int
LCGetAveragedRay(FILE *ceilfile,LCTime *time,LCCloudBases *cloudbases,
		 LCBackscatterProfile *aveprofile,int nrays)
{
  int loop=0,status=0,i;
  LCBackscatterProfile profile;
  LCCloudBases null_cloudbases;
  LCTime null_time;
  int sumbeta[LC_MAX_GATES]; /* can't use a short int - too small */
  float sumelev = 0;

  /* Skip corrupted rays */
  while (LCGetFullRay(ceilfile,time,cloudbases,aveprofile) == 0) {
    if (feof(ceilfile)) {
      return 0;
    }
  }
  for (i=0;i<aveprofile->ngates;i++) {
    sumbeta[i] = aveprofile->beta[i];
  }
  sumelev = aveprofile->elevation;

  loop=1;
  while (loop<nrays && status==0) {
    if (LCGetFullRay(ceilfile,&null_time,&null_cloudbases,&profile) > 0) {
      if (profile.ngates==aveprofile->ngates) {
	for (i=0;i<aveprofile->ngates;i++) {
	  sumbeta[i] += profile.beta[i];
	}
	sumelev += profile.elevation;
	loop++;
      }
      else status=-1;
    }
    else status=1;
  }
  if (loop > 1) {
    for (i=0;i<aveprofile->ngates;i++) {
      aveprofile->beta[i] = sumbeta[i]/loop; /* remember that beta is an integer! */
    }
    aveprofile->elevation = sumelev/loop;
  }

  if (status==0) status=1;
  return loop*status;
}

/* Average a ray in the vertical */
int
LCAverageRay(LCBackscatterProfile * profile_in,
	     LCBackscatterProfile * profile_out,int nave)
{ 
  int i,j;
  int sumbeta;
  int sumrange;

  profile_out->elevation = profile_in->elevation;
  profile_out->ngates = profile_in->ngates / nave;
  for (i=0;i<profile_out->ngates;i++) {
    sumbeta=0;
    sumrange=0;
    for (j=0;j<nave;j++) {
      sumbeta += profile_in->beta[i*nave+j];
      sumrange += profile_in->range[i*nave+j];
    }
    profile_out->beta[i] = sumbeta / nave;
    profile_out->range[i] = sumrange / nave;
  }
  return 0;
}
